#!/usr/bin/env python
# coding: utf-8

# # SOFTWARE DOCUMENTATION OF THE PLANNING TOOL
# 
# 
# # GENERAL
# 
# **Project:** Het COVID-19 Openbaar Vervoer Capaciteitsmodel: een beleidsondersteunend instrument voor de optimalisatie van de coronacapaciteit van het openbaar vervoer  
# 
# **Grant number:** 10430042010018 (50-56300-98-1606)
# 
# **Copyright:** Dr Konstantinos Gkiotsalitis, Assistant Professor, University of Twente.
# 
# **Address:** University of Twente
# Faculty of Engineering Technology
# Horst Complex (building no. 20), room Z222
# De Horst 2
# 7522LW  Enschede
# The Netherlands
# 
# **Contact:** k.gkiotsalitis@utwente.nl
# 
# **Version:** v1
# 
# 
# # PROJECT SCOPE
# 
# Public transport has a vital function within the Dutch society. To minimize the spread of the COVID-19 in public transport, a protocol has been drawn up, including rules for the maximum occupancy of buses and trains. According to the protocol, from 1 June 2020, the capacity will be about 40% for the train and 30-40% for bus, tram and metro, depending on the type of vehicle. This has recently changed because of the second COVID-19 wave. 
# 
# This project devises **public transport planning tools** to comply with the **pandemic-imposed capacity limit** (which we call “corona-capacity”). The imposed restrictions in public transport have an extreme influence on passenger behavior, turnover and service performance of carriers. The extent to which public transport companies succeed in offering safe and efficient public transport during the COVID-19 crisis has a major influence on the possible spread of the COVID-19 virus, the functioning of society (facilitating, among other things, commuting and business transport) and the costs of public transport. The current public transport planning models focus on optimizing network efficiency: that is, balancing the possible operational costs for the carrier against the quality of service for the traveler. These public transport planning models are not applicable during the COVID-19 crisis, in which we should also ensure sufficient social distancing in public transport. 
# 
# In this project we develop a **new public transport model that can map the consequences of the corona capacity** and offer **service planning advice**, based on OV-chipkaart data and scenario analysis. The application is aimed at the Keolis public transport network in Regio Twente. Regio Twente has a complex public transport network of bus and train connections (Zutphen-Oldenzaal; Enschede-Zwolle) that serve both urban and non-urban areas. 
# 
# In this documentation we document the software code for our public transport planning and demand management model that considers the regulations regarding the pandemic-imposed capacity. The model is programmed in an abstract form to increase interoperability (e.g., it can be used by train services or bus services). At the tactical planning level, this model considers the available public transport resources (vehicles) in the study area, the pandemic-imposed capacity limit (which varies over time due to the ever-shifting regulations), the passenger demand per different passenger group, and the demand management possilities per user group. Its aim is to: 
# 1.	**Plan the optimal service frequencies** while considering the pandemic-imposed capacity and the associated revenue losses
# 2.	**Incorporate demand management strategies** for specific passenger groups
# 3.	**Calculate the revenue losses** for the public transport operator due to the pandemic-imposed capacity that results in lower vehicle occupancies
# 
#         

# # DOCUMENTATION
# 
# 
# ## MODEL ASSUMPTIONS
# 
# 1. The passenger arrival rate is stable within each 1-hour period of the day. That is, passenger arrivals at stations are uniformly distributed within the 1-hour period.
# 2. The passenger demand in the COVID-19 era is inelastic to the changes made in service frequencies because the considerable demand drop is mainly an effect of travelers preferring private modes of transportation. 
# 3. The bus lines in the network do not share the same corridors.
# 4. Passengers act rationally and maintain the longest distance possible with each other when they are inside a public transport vehicle.
#
# 
# ## MODEL FORMULATION
# 
# The formulated mathematical model is casted as follows:
# 
# \begin{align}
# \min~ & W\sum_{l\in L}x_l + \sum_{l\in L}\sum_{\substack{s\in S_l\\s\neq |S_l|}}\sum_{\substack{y\in S_l\\y>s}}\sum_{r\in R} (F_r+M_r d_{l,sy}) \tilde{b}_{l,r,sy}\label{eq14}\\
# \text{s.t.}~ & \sum_{l\in L}x_l\leq N\label{eq15}\\
# & h_lx_l\geq T_l,\forall l\in L\label{eq16}\\
# &  h_{l}^{min}\leq h_l\leq h_{l}^{max}  ,\forall l\in L\label{eq17}\\
# & \gamma_{l,s}\leq k_l,\forall l\in L,~\forall s\in S_l\label{eq18}\\
# & \gamma_{l,1}=\sum_{y\in S_l}\sum_{r\in R}b_{l,r,1y}h_l,\forall l\in L\label{eq19}\\
# & \gamma_{l,s}=\gamma_{l,s-1}-\sum_{\substack{y\in S_l\\ y<s}}b_{l,ys}h_l+\sum_{\substack{y\in S_l\\ y>s}}b_{l,sy}h_l,\forall l\in L,~\forall s\in S_l\setminus\{1\}\label{eq20}\\
# & b_{l,r,sy}=B_{l,r,sy}-\tilde{b}_{l,r,sy},\forall l\in L,~\forall r\in R,~\forall s\in S_l,~\forall y\in S_l~|~y\geq s\label{eq21}\\
# & b_{l,sy}=\sum_{r\in R}b_{l,r,sy},\forall l\in L,~\forall s\in S_l,~\forall y\in S_l~|~y\geq s\label{eq22}\\
# & x_l\in\mathbb{Z}_{\geq 0},\forall l\in L\label{eq24}\\
# & h_l\in \mathbb{R}_{\geq 0},\forall l\in L\label{eq25}
# \end{align}

# ## SOFTWARE CODE
# 
# The code of the software is provided below. Note that the code is written in an abstract form to allow other public transport operators to use it. The data values are read from external files (either .xlsx or .txt). We offer these files for a specific scenario, but public transport operators can exchange the data files when applying this model to their own data.
# 
# The software code is developed in *Python 3*. Python is an interpreted, high-level and general-purpose programming language. Python is open source. The optimization part is performed with the optimization solver Gurobi using its Python library *gurobipy*. Gurobi supports linear and quadratic optimization in continuous and integer variables.
# 
# ### Import Python libraries
# 
# First, we import the library of the gurobi solver (`gurobipy`) and libraries for reading data from external files (`pandas`, `numpy`).

# In[1123]:


#import libraries
import os, sys
print(sys.executable) # works this time
print(sys.version)
print(sys.version_info)
import gurobipy as gp #import gurobipy library in Python as gp
from gurobipy import GRB
import pandas as pd #import pandas library as pd. It offers data structures and operations for manipulating numerical tables and time series
import numpy as np #import numpy library. It adds support for large, multi-dimensional arrays and matrices
import os #provides functions for interacting with the operating system
import ast #library that processes trees of the Python abstract syntax grammar


# We check the version of the gurobi solver and the gurobipy library. In this implementation, **Gurobi 9.03** is used.

# In[1124]:


print(gp.gurobi.version())


# We initialize our mathematical model in Gurobi

# In[1125]:


#Initialize the Gurobi model
model = gp.Model()
model.Params.OutputFlag = 0
model.Params.LogToConsole = 0
model.setParam("OutputFlag", 0)


# ### Set the Parameter values by reading data from files

# Read the number of network lines:

# In[1126]:


#read the number of the network lines from a .txt file as an integer number (dtype='int')
L = np.loadtxt('Data/lines.txt',dtype='int') 
L = np.arange(1,L+1)
L = tuple(L)


# Read the vehicle capacities $ c_l,~\forall l\in L $ for each public transport line:

# In[1127]:


#read the vehicle capacities for different lines from a .txt file as an integer number (dtype='int')
c = np.loadtxt('Data/capacities.txt',dtype='int')
c = tuple(c)
c = {i:c[i-1] for i in L}


# Read the cost, $W$, of deploying an extra public transport vehicle (e.g,. train or bus, depending on the scenario)

# In[1128]:


#read the cost of deploying an extra public transport vehicle from a .txt file
W = np.loadtxt('Data/W.txt',dtype='float') 


# Read the maximum allowed in-vehicle passenger load, $k_l$, to conform with the pandemic-imposed capacity regulations. This can differ from line to line and can change over time to comply with the new regulations.

# In[1129]:


#read the in-vehicle passenger load for the vehicles of each line to conform with the pandemic-imposed capacity limit
k = np.loadtxt('Data/lines_passengerload.txt',dtype='int') 
k = tuple(k)
k = {i:k[i-1] for i in L}


# Read the number of passenger group types, $R$

# In[1130]:


#read the number of different passenger group types from a .txt file
R = np.loadtxt('Data/R.txt',dtype='int') 


# Read the lost revenue, $F_r$, for a passenger who is refused service

# In[1131]:


# Read the lost revenue for a passenger who is refused service
file = open("Data/Fr.txt", "r")
contents = file.read(); Fr = ast.literal_eval(contents); file.close()


# Read the lost revenue per km, $M_r$, for a passenger who is refused service

# In[1132]:


# Read the lost revenue per km for a passenger who is refused service
file = open("Data/Mr.txt", "r")
contents = file.read(); Mr = ast.literal_eval(contents); file.close()


# Read the number of available public transport vehicles, $N$

# In[1133]:


# Read the number of available public transport vehicles at the network level
N = np.loadtxt('Data/N.txt',dtype='int') 


# Read the number of stops served by each line.

# In[1134]:


#read the number of stops served by each line from a .txt file as an integer number (dtype='int')
S_number = np.loadtxt('Data/line_stops.txt',dtype='int') 


# Read the round-trip travel time in minutes, $T_l$, for every line $l\in L$.

# In[1135]:


#read round-trip travel times in minutes, including the layover times, for each line from a .txt file
T = np.loadtxt('Data/line_traveltimes.txt') 
T = {i:T[i-1] for i in L}


# Read the minimum and maximul allowed headway per line, $h_l^{min},h_l^{max}$, in minutes.

# In[1136]:


#read the minimum allowed headway per line (in minutes) from a .txt file
file = open("Data/h_l_min.txt", "r")
contents = file.read(); h_l_min = ast.literal_eval(contents); file.close()


# In[1137]:


#read the maximum allowed headway per line (in minutes) from a .txt file
file = open("Data/h_l_max.txt", "r")
contents = file.read(); h_l_max = ast.literal_eval(contents); file.close()


# Read the ordered stops of all lines, $S_l$.

# In[1138]:


#read the stops of each line from a .txt file in a dictionary form {('line ID', 'ordered stop number'): 'ordered stop number'}
file = open("Data/stations_numberIDs.txt", "r")
contents = file.read(); Sl = ast.literal_eval(contents); file.close()
#for s in Sl:
    #print(s[0])


# Read the passenger demand per OD-pair of each line, $B_{l,r,sy}$.

# In[1139]:


#read origin-destination demand between OD-pairs from a .txt file
file = open("Data/Blrsy.txt", "r")
contents = file.read(); Blrsy = ast.literal_eval(contents); file.close()
#print(Blrsy)

'''
for s in range(1,149):
    for y in range(1,149):
        if y>s:
            Blrsy[1,5,s,y]=0*Blrsy[1,5,s,y]
            Blrsy[1,6,s,y]=0*Blrsy[1,6,s,y]
            Blrsy[1,3,s,y]=0*Blrsy[1,3,s,y]
            Blrsy[1,4,s,y]=0*Blrsy[1,3,s,y]
for s in range(1,81):
    for y in range(1,81):
        if y>s:
            Blrsy[2,5,s,y]=0*Blrsy[2,5,s,y] 
            Blrsy[2,6,s,y]=0*Blrsy[2,6,s,y] 
            Blrsy[2,3,s,y]=0*Blrsy[2,3,s,y] 
            Blrsy[2,4,s,y]=0*Blrsy[2,4,s,y]
'''


# Read the distance per OD-pair of each line, $d_{l,sy}$.

# In[1140]:


#read the distances between the o-d pairs of each line from a .txt file.
file = open("Data/distance_IDs.txt", "r")
contents = file.read(); d = ast.literal_eval(contents); file.close()


# ### Introduce the Model Variables
# 
# Set the variable of vehicles per line, $x_l$. Each line should have at least one vehicle to be operational.

# In[1141]:


#Initialize variable x_l denoting the number of vehicles assigned to each line l
x = model.addVars(L,vtype=gp.GRB.INTEGER, lb=1, name='x')


# Set the variable of line headways, $h_l$. The time headways of the lines cannot be negative and are expressed in minutes.

# In[1142]:


#initialize variable h_l denoting the time headway of line l in minutes.
h = model.addVars(L,vtype=gp.GRB.CONTINUOUS, lb=0, name='h')


# Set veriable $\gamma_{l,s}$ denoting the passenger load of each vehicle of a line $l$ when traveling from stop $s$ to $s+1$. The passenger load cannot be negative and, ideally, it should be lower than or equal to the maximum allowed passenger load to conform with the social distancing regulations.

# In[1143]:


#initialize variable gamma_{l,s} denoting the passenger load of each vehicle serving line l
gamma = model.addVars(L,Sl,vtype=gp.GRB.CONTINUOUS, lb=0, name='gamma')


# Set variable $b_{l,r,s,y}$ denoting the hourly passenger demand between stations $s$ and $y$ of line $l\in L$ that can be served by line $l$ for passenger type $r$ while conforming to the pandemic-imposed capacity limit.

# In[1144]:


#Initialize variable b_{l,r,s,y} of the hourly passenger demand between stations s and y of line l for passenger type r that can be served by line l while conforming to the pandemic-imposed capacity limit
b = model.addVars(L,R,Sl,Sl,vtype=gp.GRB.CONTINUOUS, lb=0, name='blrsy')


# Set variable $b_{l,s,y}$ denoting the hourly passenger demand between stations $s$ and $y$ of line $l\in L$ that can be served by line $l$ while conforming to the pandemic-imposed capacity limit.

# In[1145]:


#Initialize variable b_{l,s,y} of the hourly passenger demand between stations s and y of line l that can be served by line l while conforming to the pandemic-imposed capacity limit
bsy = model.addVars(L,Sl,Sl,vtype=gp.GRB.CONTINUOUS, lb=0, name='bsy')


# Initialize variable $\tilde{b}_{l,r,s,y}$ of the hourly passenger demand between stations $s$ and $y$ of line $l$ that cannot be accommodated by the public transport network due to the social distancing requirements.

# In[1146]:


#Initialize variable btilde_{l,r,s,y} of the hourly passenger demand between stations s and y of line l for passenger type r that cannot be accommodated by the public transport network due to the social distancing requirements
btilde = model.addVars(L,R,Sl,Sl,vtype=gp.GRB.CONTINUOUS, lb=0, name='btildelrsy')


# In[1147]:


btilde_s = model.addVars(L,Sl,vtype=gp.GRB.CONTINUOUS, lb=0, name='btildels')


# Initialize a variable that it will compute the overall number of unserved passengers.

# In[1148]:


btilde_overall = model.addVar(vtype=gp.GRB.CONTINUOUS, lb=0, name='btildel_overall')


# ### Add Constraints
# 
# Add constraint $\sum\limits_{l\in L}x_l\leq N$ that does not allow to use more vehicles than the maximum number of available vehicles, $N$.

# In[1149]:


model.addConstr(sum(x[l] for l in L) <= N)


# Add constraints to ensure that the operating headway is within the limits of the minimum and the maximum possible headway.

# In[1150]:


model.addConstrs(h[l]<=h_l_max[l] for l in L)
model.addConstrs(h[l]>=h_l_min[l] for l in L)


# Add constraints $h_lx_l \geq T_l,~~\forall l\in L$.

# In[1151]:


model.addConstrs( h[l]*x[l] >= T[l] for l in L ) 


# Add constraints $\gamma_{l,s}\leq k_l,~~\forall l\in L, \forall s\in S_l$

# In[1152]:


model.addConstrs(gamma[l,s[0],s[1]] <= k[l] for l in L for s in Sl if s[0]==l)


# Add constraints $\gamma_{l,1}=\sum_{y\in S_l} b_{l,1,y}h_l,~~\forall l\in L$

# In[1153]:


model.addConstrs(gamma[l,l,1] == 0.01667*h[l]*sum(sum(b[l,r,l,1,y[0],y[1]] for y in Sl if y[0]==l and y[1]>1) for r in R) for l in L)


# Add constraints $\gamma_{l,s}=\gamma_{l,s-1}-\sum\limits_{y\in S_l~|~y<s}b_{l,ys}h_l+\sum\limits_{y\in S_l~|~y>s}b_{l,sy}h_l~~(~\forall l\in L,~\forall s\in S_l\setminus\{1\}~)$

# In[1154]:


model.addConstrs(gamma[l,s[0],s[1]] == gamma[l,s[0],s[1]-1] - 
                 sum(bsy[l,y[0],y[1],s[0],s[1]]*(1/60)*h[l] for y in Sl if y[0]==l and y[1]<s[1]) + 
                 sum(bsy[l,s[0],s[1],y[0],y[1]]*(1/60)*h[l] for y in Sl if y[0]==l and y[1]>s[1]) for l in L for s in Sl if s[0]==l and s[1]!=1)


# Add constraint $b_{l,sy}=\sum_{r\in R}b_{l,r,sy},\forall l\in L,~\forall s\in S_l,~\forall y\in S_l~|~y\geq s$.

# In[1155]:


model.addConstrs(bsy[l,y[0],y[1],s[0],s[1]]==sum(b[l,r,y[0],y[1],s[0],s[1]] for r in R) for l in L for y in Sl if y[0]==l for s in Sl if s[0]==l and s[1]>y[1])


# In[1156]:


model.addConstrs(btilde_s[l,y[0],y[1]]==sum(sum(btilde[l,r,y[0],y[1],s[0],s[1]] for s in Sl if s[0]==l and s[1]>y[1])for r in R) for l in L for y in Sl if y[0]==l)


# In[1157]:


model.addConstr(btilde_overall==sum(sum(btilde_s[l,y[0],y[1]] for y in Sl if y[0]==l) for l in L))


# Add constraint $b_{l,r,sy}=B_{l,r,sy}-\tilde{b}_{l,r,sy}~~(~\forall l\in L,r\in R, s\in S_l, y\in S_l~|~y>s~)$

# In[1158]:


model.addConstrs(b[l,r,s[0],s[1],y[0],y[1]] == Blrsy[l,r,s[1],y[1]]-btilde[l,r,s[0],s[1],y[0],y[1]] 
                 for l in L for r in R for s in Sl if s[0]==l for y in Sl if y[0]==l and y[1]>s[1])


# ### Set the Objective Function
# 
# $z(x,h):= W\sum\limits_{l\in L}x_l +  \sum\limits_{l\in L}\sum\limits_{r\in R}\sum\limits_{s\in S_l\setminus\{|S_l|\}}\sum\limits_{y\in S_l~|~y>s} \big(F_r+M_r d_{l,sy}\big)\tilde{b}_{l,r,sy}$

# In[1159]:


#Declare objective function
obj = W*sum(x[l] for l in L) + sum(sum(sum(sum( (Fr[r]+Mr[r]*d[l,s[1],y[1]])*btilde[l,r,s[0],s[1],y[0],y[1]] for y in Sl if y[0]==l and y[1]>s[1]) for s in Sl if s[0]==l) for r in R) for l in L)

#Add objective function to model and declare that we solve a minimization problem
model.setObjective(obj,GRB.MINIMIZE)


# Solve the model and return results.

# In[1160]:


model.params.NonConvex = 2 #allow to handle quadratic equality constraints - which are always non-convex
model.optimize()
if model.status == GRB.OPTIMAL: #check if the solver is capable of finding an optimal solution
    model.printAttr('X')
    print(model.status,'optimal')
    print('Obj: %g' % model.objVal)
else:
    print(model.status,'not optimal')
    
#print results
for v in model.getVars():
    if v.x>0:
        print('%s %g' % (v.varName, v.x))

